"""
Regression results class for PyRegHDFE.
"""

from typing import Optional, List, Union, Dict, Any
import numpy as np
import pandas as pd
from scipy import stats
from tabulate import tabulate


class RegressionResults:
    """
    Results class for reghdfe estimations.
    
    This class contains estimation results and provides methods for displaying
    and accessing regression outputs, similar to statsmodels results.
    
    Attributes
    ----------
    params : pd.Series
        Estimated coefficients
    bse : pd.Series
        Standard errors of coefficients
    tvalues : pd.Series
        t-statistics for coefficients
    pvalues : pd.Series
        p-values for coefficients
    nobs : int
        Number of observations
    df_resid : int
        Degrees of freedom for residuals
    df_model : int
        Degrees of freedom for model
    rsquared : float
        R-squared
    rsquared_within : float
        Within R-squared (after fixed effects)
    rsquared_adj : float
        Adjusted R-squared
    fvalue : float
        F-statistic for overall significance
    f_pvalue : float
        p-value for F-statistic
    """
    
    def __init__(
        self,
        params: np.ndarray,
        cov_matrix: np.ndarray,
        var_names: List[str],
        nobs: int,
        df_resid: int,
        df_model: int,
        rsquared: float,
        rsquared_within: float,
        rsquared_adj: float,
        fvalue: float,
        f_pvalue: float,
        y_name: str,
        fe_info: Dict[str, Any],
        cluster_info: Optional[Dict[str, Any]] = None,
        weights_info: Optional[Dict[str, Any]] = None,
        iterations: Optional[int] = None,
        converged: bool = True,
        cov_type: str = "robust",
    ) -> None:
        """
        Initialize regression results.
        
        Parameters
        ----------
        params : np.ndarray
            Coefficient estimates
        cov_matrix : np.ndarray
            Covariance matrix of coefficients
        var_names : List[str]
            Names of variables
        nobs : int
            Number of observations
        df_resid : int
            Residual degrees of freedom
        df_model : int
            Model degrees of freedom
        rsquared : float
            R-squared
        rsquared_within : float
            Within R-squared
        rsquared_adj : float
            Adjusted R-squared
        fvalue : float
            F-statistic
        f_pvalue : float
            F-statistic p-value
        y_name : str
            Name of dependent variable
        fe_info : Dict[str, Any]
            Information about fixed effects
        cluster_info : Optional[Dict[str, Any]]
            Information about clustering
        weights_info : Optional[Dict[str, Any]]
            Information about weights
        iterations : Optional[int]
            Number of iterations for HDFE solver
        converged : bool
            Whether estimation converged
        cov_type : str
            Type of covariance matrix used
        """
        
        # Store core results
        self.params = pd.Series(params, index=var_names, name="coef")
        self.bse = pd.Series(np.sqrt(np.diag(cov_matrix)), index=var_names, name="std_err")
        self.tvalues = pd.Series(self.params / self.bse, index=var_names, name="t")
        self.pvalues = pd.Series(2 * (1 - stats.t.cdf(np.abs(self.tvalues), df_resid)), 
                                index=var_names, name="P>|t|")
        
        # Store degrees of freedom for confidence interval calculation
        self._df_resid = df_resid
        
        # Store other attributes
        self.nobs = nobs
        self.df_resid = df_resid
        self.df_model = df_model
        self.rsquared = rsquared
        self.rsquared_within = rsquared_within
        self.rsquared_adj = rsquared_adj
        self.fvalue = fvalue
        self.f_pvalue = f_pvalue
        self.y_name = y_name
        self.fe_info = fe_info
        self.cluster_info = cluster_info
        self.weights_info = weights_info
        self.iterations = iterations
        self.converged = converged
        self.cov_type = cov_type
        
        # Store full covariance matrix
        self.cov_matrix = pd.DataFrame(cov_matrix, index=var_names, columns=var_names)
        
    @property
    def vcov(self) -> pd.DataFrame:
        """Variance-covariance matrix of coefficients."""
        return self.cov_matrix
    
    def conf_int(self, alpha: float = 0.05) -> pd.DataFrame:
        """
        Confidence intervals for coefficients.
        
        Parameters
        ----------
        alpha : float, optional
            Significance level. Default is 0.05 for 95% confidence intervals.
            
        Returns
        -------
        pd.DataFrame
            DataFrame with lower and upper confidence bounds
        """
        t_crit = stats.t.ppf(1 - alpha/2, self._df_resid)
        lower_bound = self.params - t_crit * self.bse
        upper_bound = self.params + t_crit * self.bse
        
        return pd.DataFrame({
            f"{alpha/2:.3f}": lower_bound,
            f"{1-alpha/2:.3f}": upper_bound
        }, index=self.params.index)
        
    def summary(self, alpha: float = 0.05) -> str:
        """
        Generate summary table similar to reghdfe output.
        
        Parameters
        ----------
        alpha : float
            Significance level for confidence intervals
            
        Returns
        -------
        str
            Formatted summary table
        """
        
        lines = []
        
        # Header
        lines.append("=" * 78)
        lines.append(f"HDFE Linear regression")
        lines.append(f"Number of obs      = {self.nobs:>10,}")
        
        # Fixed effects info
        if self.fe_info:
            lines.append(f"Absorbing {len(self.fe_info['names'])} HDFE group(s)")
            
        lines.append("")
        
        # Statistics
        stats_table = [
            ["", ""],
            [f"R-squared", f"{self.rsquared:.4f}"],
            [f"Within R-sq.", f"{self.rsquared_within:.4f}"],
            [f"Adj. R-squared", f"{self.rsquared_adj:.4f}"],
            [f"F({self.df_model}, {self.df_resid})", f"{self.fvalue:.2f}"],
            [f"Prob > F", f"{self.f_pvalue:.4f}"],
        ]
        
        lines.append(tabulate(stats_table, tablefmt="plain"))
        lines.append("")
        
        # Coefficients table
        t_crit = stats.t.ppf(1 - alpha/2, self.df_resid)
        conf_low = self.params - t_crit * self.bse  
        conf_high = self.params + t_crit * self.bse
        
        coef_data = []
        coef_data.append([self.y_name, "Coef.", "Std. Err.", "t", "P>|t|", 
                         f"[{alpha:.1%} Conf. Interval]", ""])
        coef_data.append(["-" * 12, "-" * 10, "-" * 10, "-" * 8, "-" * 8, 
                         "-" * 13, "-" * 13])
        
        for var in self.params.index:
            coef_data.append([
                var,
                f"{self.params[var]:10.6f}",
                f"{self.bse[var]:10.6f}",
                f"{self.tvalues[var]:8.2f}",
                f"{self.pvalues[var]:8.3f}",
                f"{conf_low[var]:10.6f}",
                f"{conf_high[var]:10.6f}"
            ])
            
        lines.append(tabulate(coef_data, tablefmt="plain"))
        lines.append("")
        
        # Fixed effects and clustering info
        if self.fe_info:
            lines.append("Fixed effects:")
            for i, (name, info) in enumerate(zip(self.fe_info['names'], self.fe_info['counts'])):
                lines.append(f"  {name}: {info} groups")
                
        if self.cluster_info:
            lines.append(f"\nClustered standard errors: {self.cluster_info['method']}")
            if 'n_clusters' in self.cluster_info:
                for dim, n_clust in enumerate(self.cluster_info['n_clusters']):
                    lines.append(f"  Dimension {dim+1}: {n_clust} clusters")
                    
        if self.iterations:
            convergence_str = "converged" if self.converged else "did not converge"
            lines.append(f"\nHDFE algorithm {convergence_str} in {self.iterations} iterations")
            
        lines.append("=" * 78)
        
        return "\n".join(lines)
        
    def __repr__(self) -> str:
        """String representation."""
        return f"<PyRegHDFE Results: {len(self.params)} coefficients, {self.nobs} observations>"
        
    def __str__(self) -> str:
        """String representation using summary."""
        return self.summary()


"""
Public API for PyRegHDFE.

This module provides the main user-facing function for estimating
linear regressions with high-dimensional fixed effects.
"""

from typing import Optional, Union, List, Literal, Dict, Any
import pandas as pd

from .core import estimate_reghdfe
from .results import RegressionResults


def reghdfe(
    data: pd.DataFrame,
    y: str,
    x: Union[List[str], str],
    fe: Optional[Union[List[str], str]] = None,
    cluster: Optional[Union[List[str], str]] = None,
    weights: Optional[str] = None,
    drop_singletons: bool = True,
    absorb_tolerance: float = 1e-8,
    robust: bool = True,
    cov_type: Literal["robust", "cluster"] = "robust",
    ddof: Optional[int] = None,
    absorb_method: Optional[str] = None,
    absorb_options: Optional[Dict[str, Any]] = None
) -> RegressionResults:
    """
    Estimate linear regression with high-dimensional fixed effects.
    
    This function replicates the core functionality of Stata's reghdfe command,
    providing efficient estimation of linear models with multiple high-dimensional
    fixed effects using the pyhdfe library for fixed effect absorption.
    
    Parameters
    ----------
    data : pd.DataFrame
        Input dataset containing all variables
    y : str
        Name of the dependent variable
    x : Union[List[str], str]
        Name(s) of independent variables. Can be a single string or list of strings.
    fe : Optional[Union[List[str], str]], default None
        Name(s) of fixed effect variables. If None, no fixed effects are absorbed.
        Can be a single string or list of strings for multiple dimensions.
    cluster : Optional[Union[List[str], str]], default None
        Name(s) of cluster variables for cluster-robust standard errors.
        Supports 1-way and 2-way clustering. If None, uses robust standard errors.
    weights : Optional[str], default None
        Name of weight variable for weighted least squares. Uses frequency/analytic weights.
    drop_singletons : bool, default True
        Whether to drop singleton groups (groups with only one observation).
        Recommended to keep True for proper degrees of freedom calculation.
    absorb_tolerance : float, default 1e-8
        Convergence tolerance for iterative fixed effect absorption algorithms.
    robust : bool, default True
        Whether to use heteroskedasticity-robust standard errors when cluster is None.
    cov_type : Literal["robust", "cluster"], default "robust"
        Type of covariance matrix estimation:
        - "robust": Heteroskedasticity-robust (HC1)
        - "cluster": Cluster-robust (requires cluster variables)
    ddof : Optional[int], default None
        Degrees of freedom correction. If None, computed automatically as
        k_vars + n_absorbed_fe where k_vars is number of regressors and
        n_absorbed_fe is number of absorbed fixed effect parameters.
    absorb_method : Optional[str], default None
        Algorithm for fixed effect absorption:
        - None: Auto-select (within transform for 1 FE, MAP for multiple)
        - "within": Within transform (only for single fixed effect)
        - "map": Method of alternating projections
        - "lsmr": LSMR sparse solver
        - "sw": Somaini-Wolak method (only for 2 fixed effects)
    absorb_options : Optional[Dict[str, Any]], default None
        Additional options passed to the pyhdfe absorption algorithm.
        See pyhdfe documentation for available options.
        
    Returns
    -------
    RegressionResults
        Estimation results containing coefficients, standard errors, test statistics,
        R-squared measures, and other regression diagnostics. Provides a summary()
        method for formatted output similar to Stata's reghdfe.
        
    Raises
    ------
    ValueError
        - If required variables are missing from the dataset
        - If there are insufficient degrees of freedom
        - If fixed effect variables have insufficient variation
        - If cluster variables are specified but cov_type != "cluster"
        - If weights contain non-positive values
        
    Examples
    --------
    Basic regression with firm and year fixed effects:
    
    >>> import pandas as pd
    >>> from pyreghdfe import reghdfe
    >>> 
    >>> # Load your data
    >>> df = pd.read_csv("wage_data.csv")
    >>> 
    >>> # Estimate model
    >>> results = reghdfe(
    ...     data=df,
    ...     y="log_wage",
    ...     x=["experience", "education", "tenure"],
    ...     fe=["firm_id", "year"],
    ...     cluster="firm_id"
    ... )
    >>> 
    >>> # Display results
    >>> print(results.summary())
    >>> 
    >>> # Access coefficients
    >>> print(results.params)
    >>> print(results.bse)
    
    Regression with weights and custom absorption options:
    
    >>> results = reghdfe(
    ...     data=df,
    ...     y="outcome",
    ...     x=["treatment", "control_var"],
    ...     fe=["state", "year"],
    ...     weights="survey_weight",
    ...     absorb_method="lsmr",
    ...     absorb_options={"tol": 1e-10, "iteration_limit": 5000}
    ... )
    
    Two-way clustered standard errors:
    
    >>> results = reghdfe(
    ...     data=df,
    ...     y="returns",
    ...     x=["beta", "size", "momentum"],
    ...     fe=["firm", "date"],
    ...     cluster=["firm", "industry"],
    ...     cov_type="cluster"
    ... )
        
    Notes
    -----
    This implementation follows the algorithms described in:
    
    - Correia, S. (2017). Linear Models with High-Dimensional Fixed Effects: 
      An Efficient and Feasible Estimator. Working Paper.
    - Guimarães, P. and Portugal, P. (2010). A simple approach to quantify the bias
      of estimators in non-linear panel models. Journal of Econometrics, 157(2), 334-344.
      
    The function uses the Frisch-Waugh-Lovell theorem to partial out fixed effects
    before running OLS, which is computationally efficient for high-dimensional problems.
    
    Standard error calculations follow:
    - Robust: HC1 heteroskedasticity-robust (White/Huber-White)
    - Cluster: Liang-Zeger cluster-robust with small-sample corrections
    - Two-way cluster: Cameron-Gelbach-Miller method
    
    See Also
    --------
    RegressionResults : Results class with summary methods and diagnostics
    """
    
    # Input validation
    if not isinstance(data, pd.DataFrame):
        raise TypeError("data must be a pandas DataFrame")
        
    if data.empty:
        raise ValueError("data cannot be empty")
        
    # Validate covariance type and cluster consistency
    if cov_type == "cluster" and cluster is None:
        raise ValueError("cluster variables must be specified when cov_type='cluster'")
        
    if cluster is not None and cov_type != "cluster":
        # Auto-switch to cluster covariance if cluster vars provided
        cov_type = "cluster"
        
    # Validate absorb_method for fixed effects dimensions
    if absorb_method is not None:
        valid_methods = ["within", "map", "lsmr", "sw", "dummy"]
        if absorb_method not in valid_methods:
            raise ValueError(f"absorb_method must be one of {valid_methods}")
            
        # Check method compatibility with FE dimensions
        if fe is not None:
            fe_vars = [fe] if isinstance(fe, str) else fe
            n_fe_dims = len(fe_vars)
            
            if absorb_method == "within" and n_fe_dims > 1:
                raise ValueError("within method only supports single fixed effect")
            elif absorb_method == "sw" and n_fe_dims != 2:
                raise ValueError("sw method only supports exactly 2 fixed effects")
                
    # Validate cluster dimensions
    if cluster is not None:
        cluster_vars = [cluster] if isinstance(cluster, str) else cluster
        if len(cluster_vars) > 2:
            raise ValueError("Only 1-way and 2-way clustering are supported")
            
    # Call core estimation function
    try:
        results = estimate_reghdfe(
            data=data,
            y=y,
            x=x,
            fe=fe,
            cluster=cluster,
            weights=weights,
            drop_singletons=drop_singletons,
            absorb_tolerance=absorb_tolerance,
            robust=robust,
            cov_type=cov_type,
            ddof=ddof,
            absorb_method=absorb_method,
            absorb_options=absorb_options
        )
        
        return results
        
    except Exception as e:
        # Re-raise with more context
        raise ValueError(f"Estimation failed: {str(e)}") from e


# Convenience alias for backwards compatibility and alternative naming
hdfe_regression = reghdfe


"""
Tests for covariance matrix estimation.
"""

import pytest
import numpy as np
import pandas as pd
from pyreghdfe import reghdfe
from pyreghdfe.covariance import robust_covariance, cluster_covariance


class TestCovarianceMatrices:
    """Test different covariance matrix estimators."""
    
    def test_robust_covariance_hc1(self):
        """Test HC1 robust covariance matrix."""
        np.random.seed(42)
        n, k = 100, 3
        
        # Generate heteroskedastic data
        X = np.random.normal(0, 1, (n, k))
        error_variance = 0.1 + 0.5 * X[:, 0]**2  # Heteroskedasticity
        residuals = np.random.normal(0, np.sqrt(error_variance))
        
        # Compute robust covariance
        cov_robust = robust_covariance(X, residuals, hc_type="HC1")
        
        # Check dimensions
        assert cov_robust.shape == (k, k)
        
        # Check symmetry
        np.testing.assert_allclose(cov_robust, cov_robust.T)
        
        # Check positive definite
        eigenvals = np.linalg.eigvals(cov_robust)
        assert np.all(eigenvals > 0)
        
    def test_cluster_covariance_oneway(self):
        """Test one-way cluster-robust covariance."""
        np.random.seed(123)
        n, k = 200, 2
        n_clusters = 20
        
        # Generate clustered data
        X = np.random.normal(0, 1, (n, k))
        cluster_ids = np.random.randint(0, n_clusters, n)
        
        # Create cluster correlation in residuals
        cluster_effects = np.random.normal(0, 0.5, n_clusters)
        residuals = (cluster_effects[cluster_ids] + 
                    np.random.normal(0, 0.2, n))
        
        # Compute cluster-robust covariance
        cov_cluster, cluster_info = cluster_covariance(X, residuals, cluster_ids)
        
        # Check dimensions and properties
        assert cov_cluster.shape == (k, k)
        np.testing.assert_allclose(cov_cluster, cov_cluster.T)
        
        # Check cluster info
        assert cluster_info['method'] == '1-way clustering'
        assert len(cluster_info['n_clusters']) == 1
        assert cluster_info['n_clusters'][0] == n_clusters
        
    def test_cluster_covariance_twoway(self):
        """Test two-way cluster-robust covariance."""
        np.random.seed(456)
        n, k = 150, 2
        n_clusters1, n_clusters2 = 15, 10
        
        # Generate data with two clustering dimensions
        X = np.random.normal(0, 1, (n, k))
        cluster1_ids = np.random.randint(0, n_clusters1, n)
        cluster2_ids = np.random.randint(0, n_clusters2, n)
        
        # Create two-way cluster correlation
        cluster1_effects = np.random.normal(0, 0.3, n_clusters1)
        cluster2_effects = np.random.normal(0, 0.3, n_clusters2)
        residuals = (cluster1_effects[cluster1_ids] + 
                    cluster2_effects[cluster2_ids] + 
                    np.random.normal(0, 0.1, n))
        
        # Compute two-way cluster-robust covariance
        cov_cluster, cluster_info = cluster_covariance(
            X, residuals, [cluster1_ids, cluster2_ids]
        )
        
        # Check properties
        assert cov_cluster.shape == (k, k)
        np.testing.assert_allclose(cov_cluster, cov_cluster.T)
        
        # Check cluster info
        assert '2-way' in cluster_info['method']
        assert len(cluster_info['n_clusters']) == 2
        assert cluster_info['n_clusters'][0] == n_clusters1
        assert cluster_info['n_clusters'][1] == n_clusters2
        
    def test_regression_covariance_types(self):
        """Test different covariance types in full regression."""
        # Generate data with clustering structure
        np.random.seed(789)
        n = 100
        n_clusters = 10
        
        data = pd.DataFrame({
            'cluster': np.random.randint(0, n_clusters, n),
            'x1': np.random.normal(0, 1, n),
            'x2': np.random.normal(0, 1, n)
        })
        
        # Add cluster effects to residuals
        cluster_effects = np.random.normal(0, 0.5, n_clusters)
        data['cluster_effect'] = data['cluster'].map(dict(enumerate(cluster_effects)))
        data['y'] = (0.5 * data['x1'] - 0.3 * data['x2'] + 
                    data['cluster_effect'] + 
                    np.random.normal(0, 0.2, n))
        
        # Test different covariance types
        results_robust = reghdfe(
            data=data,
            y='y',
            x=['x1', 'x2'],
            cov_type='robust'
        )
        
        results_cluster = reghdfe(
            data=data,
            y='y',
            x=['x1', 'x2'],
            cluster='cluster',
            cov_type='cluster'
        )
        
        # Coefficients should be the same
        np.testing.assert_allclose(results_robust.params, results_cluster.params)
        
        # Standard errors should generally be different
        # (though not always, depending on the data)
        se_diff = np.abs(results_robust.bse - results_cluster.bse)
        
        # Check that cluster adjustment is applied
        assert results_cluster.cluster_info is not None
        assert results_robust.cluster_info is None
        
    def test_weighted_covariance(self):
        """Test covariance matrices with weights."""
        np.random.seed(101)
        n, k = 80, 2
        
        # Generate data with weights
        X = np.random.normal(0, 1, (n, k))
        weights = np.random.uniform(0.5, 2.0, n)
        residuals = np.random.normal(0, 1, n)
        
        # Test robust covariance with weights
        cov_weighted = robust_covariance(X, residuals, weights, hc_type="HC1")
        cov_unweighted = robust_covariance(X, residuals, hc_type="HC1")
        
        # Should be different
        assert not np.allclose(cov_weighted, cov_unweighted)
        
        # Both should be positive definite
        assert np.all(np.linalg.eigvals(cov_weighted) > 0)
        assert np.all(np.linalg.eigvals(cov_unweighted) > 0)
        
    def test_hc_types(self):
        """Test different heteroskedasticity correction types."""
        np.random.seed(202)
        n, k = 50, 2
        
        X = np.random.normal(0, 1, (n, k))
        residuals = np.random.normal(0, 1, n)
        
        # Test different HC types
        hc_types = ["HC0", "HC1", "HC2", "HC3"]
        cov_matrices = {}
        
        for hc_type in hc_types:
            cov_matrices[hc_type] = robust_covariance(X, residuals, hc_type=hc_type)
            
        # All should be positive definite
        for hc_type, cov in cov_matrices.items():
            eigenvals = np.linalg.eigvals(cov)
            assert np.all(eigenvals > 0), f"{hc_type} covariance not positive definite"
        
        # HC1 should be larger than HC0 (has df adjustment)
        assert np.all(np.diag(cov_matrices["HC1"]) >= np.diag(cov_matrices["HC0"]))
        
    def test_small_sample_corrections(self):
        """Test small sample corrections for cluster covariance."""
        np.random.seed(303)
        n, k = 30, 2
        n_clusters = 5  # Small number of clusters
        
        X = np.random.normal(0, 1, (n, k))
        cluster_ids = np.random.randint(0, n_clusters, n)
        residuals = np.random.normal(0, 1, n)
        
        # Test with and without small sample correction
        cov_corrected, _ = cluster_covariance(X, residuals, cluster_ids, small_sample=True)
        cov_uncorrected, _ = cluster_covariance(X, residuals, cluster_ids, small_sample=False)
        
        # Corrected should generally be larger
        # (though this is data-dependent)
        assert np.all(np.diag(cov_corrected) >= np.diag(cov_uncorrected) * 0.9)
        
    def test_numerical_stability(self):
        """Test numerical stability with near-singular designs."""
        np.random.seed(404)
        n, k = 100, 3
        
        # Create nearly collinear design
        X = np.random.normal(0, 1, (n, k))
        X[:, 2] = X[:, 0] + X[:, 1] + np.random.normal(0, 1e-10, n)  # Nearly collinear
        
        residuals = np.random.normal(0, 1, n)
        
        # Should not crash, though results may be unstable
        try:
            cov_robust = robust_covariance(X, residuals, hc_type="HC1")
            # If it succeeds, check basic properties
            assert cov_robust.shape == (k, k)
            np.testing.assert_allclose(cov_robust, cov_robust.T)
        except np.linalg.LinAlgError:
            # Expected for singular matrices
            pytest.skip("Design matrix is singular as expected")
            
    def test_edge_cases(self):
        """Test edge cases and error conditions."""
        n, k = 10, 2
        X = np.random.normal(0, 1, (n, k))
        residuals = np.random.normal(0, 1, n)
        
        # Test invalid HC type
        with pytest.raises(ValueError, match="Unknown HC type"):
            robust_covariance(X, residuals, hc_type="HC99")
            
        # Test mismatched dimensions
        with pytest.raises((ValueError, IndexError)):
            robust_covariance(X, residuals[:5])  # Wrong length residuals
            
        # Test single cluster (should work but may be unstable)
        cluster_ids = np.zeros(n, dtype=int)  # All same cluster
        try:
            cov_cluster, _ = cluster_covariance(X, residuals, cluster_ids)
            assert cov_cluster.shape == (k, k)
        except (np.linalg.LinAlgError, ZeroDivisionError):
            # May fail due to insufficient variation
            pass


"""
Core regression estimation for PyRegHDFE.

This module implements the main regression functionality including
OLS estimation, R-squared calculations, and degrees of freedom corrections.
"""

from typing import Optional, Union, List, Tuple, Dict, Any, Literal
import numpy as np
import pandas as pd
from scipy import stats

from .hdfe import HDFEAbsorber, prepare_fixed_effects, check_fe_variation
from .covariance import robust_covariance, cluster_covariance
from .results import RegressionResults


def estimate_reghdfe(
    data: pd.DataFrame,
    y: str,
    x: Union[str, List[str]],
    fe: Optional[Union[str, List[str]]] = None,
    cluster: Optional[Union[str, List[str]]] = None,
    weights: Optional[str] = None,
    drop_singletons: bool = True,
    absorb_tolerance: float = 1e-8,
    robust: bool = True,
    cov_type: Literal["robust", "cluster"] = "robust",
    ddof: Optional[int] = None,
    absorb_method: Optional[str] = None,
    absorb_options: Optional[Dict[str, Any]] = None
) -> RegressionResults:
    """
    Estimate linear regression with high-dimensional fixed effects.
    
    This is the core estimation function that implements the full reghdfe
    algorithm using the Frisch-Waugh-Lovell theorem.
    
    Parameters
    ----------
    data : pd.DataFrame
        Input dataset
    y : str
        Dependent variable name
    x : Union[str, List[str]]
        Independent variable names
    fe : Optional[Union[str, List[str]]]
        Fixed effect variable names
    cluster : Optional[Union[str, List[str]]]
        Cluster variable names for robust standard errors
    weights : Optional[str]
        Weight variable name
    drop_singletons : bool
        Whether to drop singleton groups
    absorb_tolerance : float
        Convergence tolerance for HDFE absorption
    robust : bool
        Whether to use heteroskedasticity-robust standard errors
    cov_type : Literal["robust", "cluster"]
        Type of covariance matrix estimation
    ddof : Optional[int]
        Degrees of freedom correction (if None, computed automatically)
    absorb_method : Optional[str]
        Method for fixed effect absorption
    absorb_options : Optional[Dict[str, Any]]
        Additional options for absorption algorithm
        
    Returns
    -------
    RegressionResults
        Estimation results
    """
    
    # Prepare variables
    y_var, x_vars, fe_vars, cluster_vars, weight_var = _prepare_variables(
        data, y, x, fe, cluster, weights
    )
    
    # Get clean data (no missing values)
    clean_data, valid_mask = _get_clean_data(data, y_var, x_vars, fe_vars, cluster_vars, weight_var)
    
    # Extract arrays
    y_array = clean_data[y_var].values.reshape(-1, 1)
    X_array = clean_data[x_vars].values
    weight_array = clean_data[weight_var].values if weight_var else None
    
    # Prepare fixed effects if specified
    hdfe_absorber = None
    if fe_vars:
        fe_ids, fe_names = prepare_fixed_effects(clean_data, fe_vars)
        check_fe_variation(fe_ids, fe_names)
        
        # Prepare cluster IDs for DoF computation
        cluster_ids_for_dof = None
        if cluster_vars:
            cluster_ids_for_dof, _ = prepare_fixed_effects(clean_data, cluster_vars)
            
        hdfe_absorber = HDFEAbsorber(
            fe_ids=fe_ids,
            cluster_ids=cluster_ids_for_dof,
            drop_singletons=drop_singletons,
            absorb_tolerance=absorb_tolerance,
            absorb_method=absorb_method,
            absorb_options=absorb_options
        )
        
        # Absorb fixed effects from y and X
        y_resid = hdfe_absorber.absorb(y_array, weight_array)
        X_resid = hdfe_absorber.absorb(X_array, weight_array)
        
        # Update valid observations after singleton dropping
        additional_mask = hdfe_absorber.valid_observations
        final_valid_mask = valid_mask.copy()
        final_valid_mask[valid_mask] = additional_mask
        
    else:
        # No fixed effects - use original data
        y_resid = y_array
        X_resid = X_array
        final_valid_mask = valid_mask
        
    # Apply weights to residualized data if specified
    if weight_var:
        if hdfe_absorber:
            # Weights already applied during absorption
            weight_final = weight_array[hdfe_absorber.valid_observations] if hdfe_absorber else weight_array
        else:
            weight_final = weight_array
            # Apply sqrt weights for WLS
            weight_sqrt = np.sqrt(weight_final).reshape(-1, 1)
            y_resid = y_resid * weight_sqrt
            X_resid = X_resid * weight_sqrt
    else:
        weight_final = None
        
    # Estimate OLS on residualized data
    coefficients, residuals = _estimate_ols(X_resid, y_resid)
    
    # Compute R-squared statistics
    r2_stats = _compute_r_squared(
        data=data[final_valid_mask],
        y_var=y_var,
        X_original=X_array if not hdfe_absorber else clean_data[x_vars].values[hdfe_absorber.valid_observations],
        y_resid=y_resid,
        X_resid=X_resid, 
        residuals=residuals,
        coefficients=coefficients,
        weights=weight_final,
        has_fe=fe_vars is not None
    )
    
    # Compute degrees of freedom
    n_obs = len(y_resid)
    k_vars = len(x_vars)
    
    if ddof is not None:
        df_model = ddof
    else:
        df_model = k_vars
        if hdfe_absorber:
            df_model += hdfe_absorber.n_absorbed_fe
            
    df_resid = n_obs - df_model
    
    if df_resid <= 0:
        raise ValueError(f"Insufficient degrees of freedom: {df_resid}")
        
    # Compute covariance matrix
    cluster_info = None
    if cov_type == "cluster" and cluster_vars:
        # Prepare cluster IDs for covariance calculation
        cluster_data = clean_data[cluster_vars]
        if hdfe_absorber:
            cluster_data = cluster_data[hdfe_absorber.valid_observations]
            
        if len(cluster_vars) == 1:
            cluster_ids_cov = cluster_data.iloc[:, 0].values
        else:
            cluster_ids_cov = [cluster_data.iloc[:, i].values for i in range(len(cluster_vars))]
            
        cov_matrix, cluster_info = cluster_covariance(
            X_resid, residuals, cluster_ids_cov, weight_final
        )
        
    elif robust or cov_type == "robust":
        cov_matrix = robust_covariance(X_resid, residuals, weight_final, hc_type="HC1")
        
    else:
        # Classical covariance matrix
        sigma2 = np.sum(residuals**2) / df_resid
        XTX_inv = np.linalg.inv(X_resid.T @ X_resid)
        cov_matrix = sigma2 * XTX_inv
        
    # Compute F-statistic
    fvalue, f_pvalue = _compute_f_statistic(coefficients, cov_matrix, df_resid)
    
    # Gather fixed effects info
    fe_info = hdfe_absorber.fe_info if hdfe_absorber else {}
    
    # Gather weights info
    weights_info = {"variable": weight_var, "type": "frequency"} if weight_var else None
    
    # Create results object
    results = RegressionResults(
        params=coefficients,
        cov_matrix=cov_matrix,
        var_names=x_vars,
        nobs=n_obs,
        df_resid=df_resid,
        df_model=df_model,
        rsquared=r2_stats["r2"],
        rsquared_within=r2_stats["r2_within"],
        rsquared_adj=r2_stats["r2_adj"],
        fvalue=fvalue,
        f_pvalue=f_pvalue,
        y_name=y_var,
        fe_info=fe_info,
        cluster_info=cluster_info,
        weights_info=weights_info,
        iterations=hdfe_absorber.iterations if hdfe_absorber else None,
        converged=hdfe_absorber.converged if hdfe_absorber else True,
        cov_type=cov_type if cov_type == "cluster" else "robust" if robust else "classical"
    )
    
    return results


def _prepare_variables(
    data: pd.DataFrame,
    y: str,
    x: Union[str, List[str]],
    fe: Optional[Union[str, List[str]]],
    cluster: Optional[Union[str, List[str]]],
    weights: Optional[str]
) -> Tuple[str, List[str], Optional[List[str]], Optional[List[str]], Optional[str]]:
    """Prepare and validate variable names."""
    
    # Convert to lists
    x_vars = [x] if isinstance(x, str) else x
    fe_vars = [fe] if isinstance(fe, str) and fe is not None else fe
    cluster_vars = [cluster] if isinstance(cluster, str) and cluster is not None else cluster
    
    # Check that all variables exist
    all_vars = [y] + x_vars
    if fe_vars:
        all_vars.extend(fe_vars)
    if cluster_vars:
        all_vars.extend(cluster_vars)
    if weights:
        all_vars.append(weights)
        
    missing_vars = [var for var in all_vars if var not in data.columns]
    if missing_vars:
        raise ValueError(f"Variables not found in data: {missing_vars}")
        
    # Check for duplicates
    if len(set(x_vars)) != len(x_vars):
        raise ValueError("Duplicate variables in x")
        
    if y in x_vars:
        raise ValueError("Dependent variable cannot be in independent variables")
        
    return y, x_vars, fe_vars, cluster_vars, weights


def _get_clean_data(
    data: pd.DataFrame,
    y_var: str,
    x_vars: List[str],
    fe_vars: Optional[List[str]],
    cluster_vars: Optional[List[str]],
    weight_var: Optional[str]
) -> Tuple[pd.DataFrame, np.ndarray]:
    """Get clean data without missing values."""
    
    # Determine all variables needed - use set to avoid duplicates
    all_vars = [y_var] + x_vars
    if fe_vars:
        all_vars.extend(fe_vars)
    if cluster_vars:
        all_vars.extend(cluster_vars)
    if weight_var:
        all_vars.append(weight_var)
    
    # Remove duplicates while preserving order
    all_vars = list(dict.fromkeys(all_vars))
        
    # Select relevant columns
    subset_data = data[all_vars].copy()
    
    # Identify complete cases
    valid_mask = subset_data.notna().all(axis=1).values
    
    if not valid_mask.any():
        raise ValueError("No complete observations found")
        
    clean_data = subset_data[valid_mask].copy()
    
    # Validate weights if provided
    if weight_var:
        weights = clean_data[weight_var]
        if (weights <= 0).any():
            raise ValueError("All weights must be positive")
            
    return clean_data, valid_mask


def _estimate_ols(X: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Estimate OLS coefficients and compute residuals."""
    
    # Check for perfect collinearity
    try:
        coefficients = np.linalg.solve(X.T @ X, X.T @ y.ravel())
    except np.linalg.LinAlgError:
        # Try with SVD for numerical stability
        try:
            U, s, Vt = np.linalg.svd(X, full_matrices=False)
            # Filter out small singular values
            s_inv = np.where(s > 1e-12, 1/s, 0)
            coefficients = Vt.T @ (s_inv[:, np.newaxis] * (U.T @ y.ravel())[:, np.newaxis]).ravel()
        except:
            raise ValueError("Design matrix is singular (perfect collinearity)")
    
    # Compute residuals
    residuals = (y.ravel() - X @ coefficients).reshape(-1, 1)
    
    return coefficients, residuals


def _compute_r_squared(
    data: pd.DataFrame,
    y_var: str,
    X_original: np.ndarray,
    y_resid: np.ndarray,
    X_resid: np.ndarray,
    residuals: np.ndarray,
    coefficients: np.ndarray,
    weights: Optional[np.ndarray],
    has_fe: bool
) -> Dict[str, float]:
    """Compute R-squared statistics."""
    
    # Get original y values
    y_original = data[y_var].values.reshape(-1, 1)
    
    # Apply weights if needed
    if weights is not None:
        weight_sqrt = np.sqrt(weights).reshape(-1, 1)
        y_original_w = y_original * weight_sqrt
        y_resid_w = y_resid
        residuals_w = residuals
    else:
        y_original_w = y_original
        y_resid_w = y_resid
        residuals_w = residuals
    
    # Total sum of squares (original y)
    if weights is not None:
        y_mean_original = np.average(y_original.ravel(), weights=weights.ravel())
        tss_original = np.sum(weights.ravel() * (y_original.ravel() - y_mean_original)**2)
    else:
        y_mean_original = np.mean(y_original)
        tss_original = np.sum((y_original - y_mean_original)**2)
    
    # Total sum of squares (after FE absorption)
    if weights is not None:
        y_mean_resid = np.average(y_resid.ravel(), weights=weights.ravel())
        tss_within = np.sum(weights.ravel() * (y_resid.ravel() - y_mean_resid)**2)
    else:
        y_mean_resid = np.mean(y_resid)
        tss_within = np.sum((y_resid - y_mean_resid)**2)
    
    # Explained sum of squares (within, after FE absorption)
    fitted_resid = X_resid @ coefficients
    if weights is not None:
        ess_within = np.sum(weights.ravel() * (fitted_resid.ravel() - y_mean_resid)**2)
    else:
        ess_within = np.sum((fitted_resid - y_mean_resid)**2)
    
    # Residual sum of squares
    if weights is not None:
        rss = np.sum(weights.ravel() * residuals_w.ravel()**2)
    else:
        rss = np.sum(residuals_w**2)
    
    # Compute R-squared measures
    r2_within = ess_within / tss_within if tss_within > 0 else 0.0
    
    if has_fe:
        # Overall R-squared: compare fitted values to original y
        fitted_original = fitted_resid  # This is the within-variation explained
        r2_overall = 1 - rss / tss_original if tss_original > 0 else 0.0
    else:
        # No fixed effects case
        r2_overall = r2_within
    
    # Adjusted R-squared  
    n = len(y_resid)
    k = len(coefficients)
    r2_adj = 1 - (1 - r2_overall) * (n - 1) / (n - k) if n > k else 0.0
    
    return {
        "r2": r2_overall,
        "r2_within": r2_within, 
        "r2_adj": r2_adj,
        "tss": tss_original,
        "ess": ess_within,
        "rss": rss
    }


def _compute_f_statistic(
    coefficients: np.ndarray,
    cov_matrix: np.ndarray,
    df_resid: int
) -> Tuple[float, float]:
    """Compute F-statistic for overall significance."""
    
    k = len(coefficients)
    
    if k == 0:
        return 0.0, 1.0
    
    try:
        # Wald test: coefficients' (V^-1) coefficients
        wald_stat = coefficients.T @ np.linalg.inv(cov_matrix) @ coefficients
        fvalue = wald_stat / k
        f_pvalue = 1 - stats.f.cdf(fvalue, k, df_resid)
        
        return float(fvalue), float(f_pvalue)
        
    except np.linalg.LinAlgError:
        # Singular covariance matrix
        return np.nan, np.nan


"""
Covariance matrix estimation for PyRegHDFE.

This module provides robust and cluster-robust covariance matrix estimators
for linear models with high-dimensional fixed effects.
"""

from typing import Optional, Union, List, Tuple, Dict, Any
import numpy as np
import pandas as pd
from scipy import sparse


def robust_covariance(
    X: np.ndarray,
    residuals: np.ndarray,
    weights: Optional[np.ndarray] = None,
    hc_type: str = "HC1"
) -> np.ndarray:
    """
    Compute heteroskedasticity-robust covariance matrix (White/Huber-White).
    
    Parameters
    ----------
    X : np.ndarray
        Design matrix (n x k)
    residuals : np.ndarray  
        Regression residuals (n x 1)
    weights : Optional[np.ndarray]
        Regression weights (n x 1)
    hc_type : str
        Type of heteroskedasticity correction: HC0, HC1, HC2, HC3
        
    Returns
    -------
    np.ndarray
        Robust covariance matrix (k x k)
    """
    
    n, k = X.shape
    
    # Apply weights if provided
    if weights is not None:
        # Ensure weights is 1D
        if weights.ndim > 1:
            weights = weights.flatten()
        
        # Apply weights: multiply each row by sqrt(weight)
        sqrt_weights = np.sqrt(weights)
        X_weighted = X * sqrt_weights.reshape(-1, 1)
        residuals_weighted = residuals.flatten() * sqrt_weights
    else:
        X_weighted = X
        residuals_weighted = residuals.flatten() if residuals.ndim > 1 else residuals
        
    # Compute (X'X)^(-1)
    XTX_inv = np.linalg.inv(X_weighted.T @ X_weighted)
    
    # Compute residual adjustments based on HC type
    if hc_type == "HC0":
        # No adjustment
        residual_adj = residuals_weighted
    elif hc_type == "HC1":
        # Degrees of freedom adjustment
        df_adj = n / (n - k)
        residual_adj = residuals_weighted * np.sqrt(df_adj)
    elif hc_type == "HC2":
        # Leverage adjustment
        H = X_weighted @ XTX_inv @ X_weighted.T
        h_ii = np.diag(H)
        residual_adj = residuals_weighted / np.sqrt(1 - h_ii)
    elif hc_type == "HC3":
        # Squared leverage adjustment  
        H = X_weighted @ XTX_inv @ X_weighted.T
        h_ii = np.diag(H)
        residual_adj = residuals_weighted / (1 - h_ii)
    else:
        raise ValueError(f"Unknown HC type: {hc_type}")
    
    # Compute meat matrix
    meat = np.zeros((k, k))
    
    for i in range(n):
        xi = X_weighted[i, :].reshape(-1, 1)  # Ensure k x 1 shape (k params x 1)
        ei = residual_adj[i]
        
        # Outer product: (k x 1) @ (1 x k) = (k x k)
        outer_product = xi @ xi.T
        
        # Add to meat matrix
        meat += ei**2 * outer_product
    
    # Sandwich formula: (X'X)^(-1) * meat * (X'X)^(-1)
    return XTX_inv @ meat @ XTX_inv


def cluster_covariance(
    X: np.ndarray,
    residuals: np.ndarray,
    cluster_ids: Union[np.ndarray, List[np.ndarray]],
    weights: Optional[np.ndarray] = None,
    small_sample: bool = True
) -> Tuple[np.ndarray, Dict[str, Any]]:
    """
    Compute cluster-robust covariance matrix (Liang-Zeger).
    
    Supports both one-way and two-way clustering.
    
    Parameters
    ----------
    X : np.ndarray
        Design matrix (n x k)
    residuals : np.ndarray
        Regression residuals (n x 1)
    cluster_ids : Union[np.ndarray, List[np.ndarray]]
        Cluster identifiers. For 2-way clustering, pass list of two arrays.
    weights : Optional[np.ndarray]
        Regression weights (n x 1)
    small_sample : bool
        Whether to apply small-sample corrections
        
    Returns
    -------
    Tuple[np.ndarray, Dict[str, Any]]
        Cluster-robust covariance matrix and cluster info
    """
    
    n, k = X.shape
    
    # Apply weights if provided
    if weights is not None:
        # Ensure weights is 1D
        if weights.ndim > 1:
            weights = weights.flatten()
        
        # Apply weights: multiply each row by sqrt(weight)
        sqrt_weights = np.sqrt(weights)
        X_weighted = X * sqrt_weights.reshape(-1, 1)
        residuals_weighted = residuals.flatten() * sqrt_weights
    else:
        X_weighted = X
        residuals_weighted = residuals.flatten() if residuals.ndim > 1 else residuals
        
    # Compute (X'X)^(-1)
    XTX_inv = np.linalg.inv(X_weighted.T @ X_weighted)
    
    # Handle single cluster dimension
    if not isinstance(cluster_ids, list):
        cluster_ids = [cluster_ids]
        
    n_cluster_dims = len(cluster_ids)
    
    if n_cluster_dims == 1:
        # One-way clustering
        cov_matrix, cluster_info = _oneway_cluster_covariance(
            X_weighted, residuals_weighted, cluster_ids[0], XTX_inv, small_sample
        )
    elif n_cluster_dims == 2:
        # Two-way clustering
        cov_matrix, cluster_info = _twoway_cluster_covariance(
            X_weighted, residuals_weighted, cluster_ids, XTX_inv, small_sample
        )
    else:
        raise ValueError("Only 1-way and 2-way clustering are supported")
        
    return cov_matrix, cluster_info


def _oneway_cluster_covariance(
    X: np.ndarray,
    residuals: np.ndarray,
    cluster_ids: np.ndarray,
    XTX_inv: np.ndarray,
    small_sample: bool = True
) -> Tuple[np.ndarray, Dict[str, Any]]:
    """Compute one-way cluster-robust covariance matrix."""
    
    n, k = X.shape
    
    # Get unique clusters and create mapping
    clusters = pd.Series(cluster_ids)
    unique_clusters = clusters.unique()
    n_clusters = len(unique_clusters)
    
    # Compute cluster sums
    meat = np.zeros((k, k))
    
    for cluster in unique_clusters:
        # Get observations in this cluster
        cluster_mask = (clusters == cluster).values
        cluster_size = cluster_mask.sum()
        
        if cluster_size == 0:
            continue
            
        # Sum residuals and regressors within cluster
        X_cluster = X[cluster_mask, :]  # nc x k
        resid_cluster = residuals[cluster_mask]  # nc x 1
        
        # Compute cluster contribution to meat matrix
        cluster_score = X_cluster.T @ resid_cluster  # k x 1
        meat += cluster_score @ cluster_score.T
    
    # Small sample correction
    if small_sample:
        # Standard correction: (G/(G-1)) * ((n-1)/(n-k))
        g_correction = n_clusters / (n_clusters - 1)
        df_correction = (n - 1) / (n - k)
        correction = g_correction * df_correction
        meat *= correction
    
    # Sandwich formula
    cov_matrix = XTX_inv @ meat @ XTX_inv
    
    cluster_info = {
        "method": "1-way clustering",
        "n_clusters": [n_clusters],
        "correction": "small-sample" if small_sample else "none"
    }
    
    return cov_matrix, cluster_info


def _twoway_cluster_covariance(
    X: np.ndarray,
    residuals: np.ndarray,
    cluster_ids: List[np.ndarray],
    XTX_inv: np.ndarray,
    small_sample: bool = True
) -> Tuple[np.ndarray, Dict[str, Any]]:
    """Compute two-way cluster-robust covariance matrix using Cameron-Gelbach-Miller."""
    
    n, k = X.shape
    cluster1_ids, cluster2_ids = cluster_ids
    
    # Compute individual cluster covariance matrices
    meat1 = _compute_cluster_meat(X, residuals, cluster1_ids)
    meat2 = _compute_cluster_meat(X, residuals, cluster2_ids)
    
    # Compute intersection clusters
    intersection_ids = np.array([f"{c1}_{c2}" for c1, c2 in zip(cluster1_ids, cluster2_ids)])
    meat_intersection = _compute_cluster_meat(X, residuals, intersection_ids)
    
    # Cameron-Gelbach-Miller formula: V1 + V2 - V_intersection
    meat = meat1 + meat2 - meat_intersection
    
    # Small sample corrections
    if small_sample:
        n_clusters1 = len(pd.Series(cluster1_ids).unique())
        n_clusters2 = len(pd.Series(cluster2_ids).unique())
        n_clusters_intersection = len(pd.Series(intersection_ids).unique())
        
        # Apply corrections to each component
        g1_correction = n_clusters1 / (n_clusters1 - 1)
        g2_correction = n_clusters2 / (n_clusters2 - 1)
        gi_correction = n_clusters_intersection / (n_clusters_intersection - 1)
        
        df_correction = (n - 1) / (n - k)
        
        # Adjust meat components
        meat = (g1_correction * df_correction * meat1 + 
                g2_correction * df_correction * meat2 - 
                gi_correction * df_correction * meat_intersection)
    
    # Sandwich formula
    cov_matrix = XTX_inv @ meat @ XTX_inv
    
    cluster_info = {
        "method": "2-way clustering (Cameron-Gelbach-Miller)",
        "n_clusters": [len(pd.Series(cluster1_ids).unique()), 
                      len(pd.Series(cluster2_ids).unique())],
        "correction": "small-sample" if small_sample else "none"
    }
    
    return cov_matrix, cluster_info


def _compute_cluster_meat(X: np.ndarray, residuals: np.ndarray, cluster_ids: np.ndarray) -> np.ndarray:
    """Compute meat matrix for a single clustering dimension."""
    
    k = X.shape[1]
    clusters = pd.Series(cluster_ids)
    unique_clusters = clusters.unique()
    
    meat = np.zeros((k, k))
    
    for cluster in unique_clusters:
        cluster_mask = (clusters == cluster).values
        
        if cluster_mask.sum() == 0:
            continue
            
        # Sum within cluster
        X_cluster = X[cluster_mask, :]
        resid_cluster = residuals[cluster_mask]
        
        # Cluster score
        cluster_score = X_cluster.T @ resid_cluster
        meat += cluster_score @ cluster_score.T
        
    return meat


"""
High-dimensional fixed effects absorption using pyhdfe.

This module provides a wrapper around the pyhdfe library to absorb
fixed effects from matrices using various algorithms.
"""

from typing import Optional, Union, List, Tuple, Dict, Any
import numpy as np
import pandas as pd
import pyhdfe


class HDFEAbsorber:
    """
    Wrapper class for absorbing high-dimensional fixed effects using pyhdfe.
    
    This class handles the interface to pyhdfe and provides methods for
    absorbing fixed effects from dependent and independent variables.
    """
    
    def __init__(
        self,
        fe_ids: np.ndarray,
        cluster_ids: Optional[np.ndarray] = None,
        drop_singletons: bool = True,
        absorb_tolerance: float = 1e-8,
        absorb_method: Optional[str] = None,
        absorb_options: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Initialize HDFE absorber.
        
        Parameters
        ----------
        fe_ids : np.ndarray
            Fixed effect identifiers (n x num_fe)
        cluster_ids : Optional[np.ndarray]
            Cluster identifiers for DoF computation
        drop_singletons : bool
            Whether to drop singleton groups
        absorb_tolerance : float
            Convergence tolerance for iterative algorithms
        absorb_method : Optional[str]
            Method for absorption ('within', 'map', 'lsmr', etc.)
        absorb_options : Optional[Dict[str, Any]]
            Additional options for pyhdfe
        """
        
        self.fe_ids = np.atleast_2d(fe_ids)
        self.cluster_ids = cluster_ids
        self.drop_singletons = drop_singletons
        self.absorb_tolerance = absorb_tolerance
        self.absorb_method = absorb_method
        self.absorb_options = absorb_options or {}
        
        # Initialize pyhdfe algorithm
        self._initialize_algorithm()
        
    def _initialize_algorithm(self) -> None:
        """Initialize the pyhdfe algorithm."""
        
        # Determine actual method that will be used by pyhdfe
        actual_method = self.absorb_method
        if actual_method is None:
            # pyhdfe auto-selects 'within' for 1 FE, 'map' for multiple
            actual_method = 'within' if self.fe_ids.shape[1] == 1 else 'map'
        
        # Set up options - only add tolerance for methods that support it
        options = self.absorb_options.copy()
        if actual_method not in ['within', 'sw', 'dummy']:
            options['tol'] = self.absorb_tolerance
        
        # Create algorithm
        try:
            self.algorithm = pyhdfe.create(
                ids=self.fe_ids,
                cluster_ids=self.cluster_ids,
                drop_singletons=self.drop_singletons,
                compute_degrees=True,
                residualize_method=self.absorb_method,
                options=options
            )
            self.converged = True
            self.iterations = getattr(self.algorithm, 'iterations_', None)
            
        except Exception as e:
            raise ValueError(f"Failed to initialize HDFE algorithm: {e}")
            
    def absorb(
        self, 
        matrices: Union[np.ndarray, List[np.ndarray]], 
        weights: Optional[np.ndarray] = None
    ) -> Union[np.ndarray, List[np.ndarray]]:
        """
        Absorb fixed effects from one or more matrices.
        
        Parameters
        ----------
        matrices : Union[np.ndarray, List[np.ndarray]]
            Matrix or list of matrices to residualize
        weights : Optional[np.ndarray]
            Observation weights
            
        Returns
        -------
        Union[np.ndarray, List[np.ndarray]]
            Residualized matrices
        """
        
        if isinstance(matrices, list):
            return [self._absorb_single(mat, weights) for mat in matrices]
        else:
            return self._absorb_single(matrices, weights)
            
    def _absorb_single(self, matrix: np.ndarray, weights: Optional[np.ndarray] = None) -> np.ndarray:
        """Absorb fixed effects from a single matrix."""
        
        # Ensure matrix is 2D
        matrix = np.atleast_2d(matrix)
        if matrix.shape[0] == 1 and matrix.shape[1] > 1:
            matrix = matrix.T
            
        # Apply weights if provided
        if weights is not None:
            weights_sqrt = np.sqrt(weights).reshape(-1, 1)
            matrix_weighted = matrix * weights_sqrt
        else:
            matrix_weighted = matrix
            
        # Remove observations dropped by singletons
        if self.algorithm.singleton_indices is not None:
            valid_obs = ~self.algorithm.singleton_indices
            matrix_weighted = matrix_weighted[valid_obs]
            
        # Residualize using pyhdfe
        try:
            residualized = self.algorithm.residualize(matrix_weighted)
            
            # Store iteration info if available
            if hasattr(self.algorithm, 'last_iterations_'):
                self.iterations = self.algorithm.last_iterations_
                
        except Exception as e:
            raise ValueError(f"Failed to absorb fixed effects: {e}")
            
        return residualized
        
    @property
    def valid_observations(self) -> np.ndarray:
        """Boolean mask of observations kept after dropping singletons."""
        if self.algorithm.singleton_indices is not None:
            return ~self.algorithm.singleton_indices
        else:
            return np.ones(self.fe_ids.shape[0], dtype=bool)
            
    @property
    def n_dropped_obs(self) -> int:
        """Number of observations dropped due to singletons."""
        if self.algorithm.singletons is not None:
            return self.algorithm.singletons
        else:
            return 0
            
    @property
    def n_absorbed_fe(self) -> int:
        """Number of fixed effect parameters absorbed."""
        if self.algorithm.degrees is not None:
            return self.algorithm.degrees
        else:
            # Fallback estimation
            return sum(len(np.unique(self.fe_ids[:, i])) - 1 
                      for i in range(self.fe_ids.shape[1]))
            
    @property
    def fe_info(self) -> Dict[str, Any]:
        """Information about absorbed fixed effects."""
        
        fe_names = [f"fe_{i+1}" for i in range(self.fe_ids.shape[1])]
        fe_counts = []
        
        valid_mask = self.valid_observations
        
        for i in range(self.fe_ids.shape[1]):
            unique_ids = np.unique(self.fe_ids[valid_mask, i])
            fe_counts.append(len(unique_ids))
            
        return {
            'names': fe_names,
            'counts': fe_counts,
            'total_absorbed': self.n_absorbed_fe,
            'method': self.absorb_method or 'auto',
            'tolerance': self.absorb_tolerance,
            'converged': self.converged,
            'iterations': self.iterations,
            'singletons_dropped': self.n_dropped_obs
        }


def prepare_fixed_effects(
    data: pd.DataFrame,
    fe_vars: Union[str, List[str]]
) -> Tuple[np.ndarray, List[str]]:
    """
    Prepare fixed effect variables for absorption.
    
    Parameters
    ----------
    data : pd.DataFrame
        Input dataset
    fe_vars : Union[str, List[str]]
        Fixed effect variable names
        
    Returns
    -------
    Tuple[np.ndarray, List[str]]
        Fixed effect IDs array and variable names
    """
    
    if isinstance(fe_vars, str):
        fe_vars = [fe_vars]
        
    # Check that all FE variables exist
    missing_vars = [var for var in fe_vars if var not in data.columns]
    if missing_vars:
        raise ValueError(f"Fixed effect variables not found: {missing_vars}")
        
    # Extract FE data and convert to numeric codes
    fe_data = []
    fe_names = []
    
    for var in fe_vars:
        fe_series = data[var].copy()
        
        # Handle missing values  
        try:
            has_nulls = pd.isna(fe_series).sum() > 0
            if has_nulls:
                raise ValueError(f"Fixed effect variable '{var}' contains missing values")
        except Exception as e:
            # If there's an issue with null checking, skip it for now
            print(f"Warning: Could not check for null values in '{var}': {e}")
            pass
            
        # Convert to categorical codes for efficiency
        if fe_series.dtype == 'object' or pd.api.types.is_categorical_dtype(fe_series):
            codes = pd.Categorical(fe_series).codes
        else:
            # For numeric variables, create a mapping
            unique_vals = fe_series.unique()
            val_to_code = {val: i for i, val in enumerate(unique_vals)}
            codes = fe_series.map(val_to_code).values
            
        fe_data.append(codes)
        fe_names.append(var)
        
    return np.column_stack(fe_data), fe_names


def check_fe_variation(fe_ids: np.ndarray, fe_names: List[str]) -> None:
    """
    Check that fixed effects have sufficient variation.
    
    Parameters
    ----------
    fe_ids : np.ndarray
        Fixed effect identifiers
    fe_names : List[str]
        Fixed effect variable names
        
    Raises
    ------
    ValueError
        If any fixed effect has insufficient variation
    """
    
    for i, name in enumerate(fe_names):
        unique_count = len(np.unique(fe_ids[:, i]))
        
        if unique_count < 2:
            raise ValueError(f"Fixed effect '{name}' has only {unique_count} unique value(s)")
            
        if i > 0 and unique_count < 2:
            raise ValueError(f"Fixed effect '{name}' (dimension {i+1}) must have at least 2 levels")


"""
PyRegHDFE: Python implementation of Stata's reghdfe for high-dimensional fixed effects regression.

This package provides efficient estimation of linear models with multiple high-dimensional fixed effects,
using the pyhdfe library for fixed effect absorption and providing robust and cluster-robust standard errors.
"""

from .api import reghdfe
from .results import RegressionResults

__version__ = "0.1.1"
__all__ = ["reghdfe", "RegressionResults", "__version__"]


"""
Basic tests for PyRegHDFE functionality.
"""

import pytest
import numpy as np
import pandas as pd
from pyreghdfe import reghdfe


class TestBasicFunctionality:
    """Test basic regression functionality."""
    
    def test_import(self):
        """Test that package imports correctly."""
        from pyreghdfe import reghdfe, RegressionResults, __version__
        assert callable(reghdfe)
        assert __version__ == "0.1.0"
        
    def test_simple_ols_no_fe(self):
        """Test OLS regression without fixed effects."""
        # Generate simple test data
        np.random.seed(42)
        n = 100
        
        data = pd.DataFrame({
            'y': np.random.normal(0, 1, n),
            'x1': np.random.normal(0, 1, n),
            'x2': np.random.normal(0, 1, n)
        })
        
        # Add true relationship
        data['y'] = 1.0 + 0.5 * data['x1'] - 0.3 * data['x2'] + np.random.normal(0, 0.1, n)
        
        # Estimate model
        results = reghdfe(
            data=data,
            y='y',
            x=['x1', 'x2']
        )
        
        # Basic checks
        assert len(results.params) == 2
        assert 'x1' in results.params.index
        assert 'x2' in results.params.index
        assert results.nobs == n
        assert results.df_resid == n - 2
        assert 0 <= results.rsquared <= 1
        
        # Check that coefficients are roughly correct
        assert abs(results.params['x1'] - 0.5) < 0.1
        assert abs(results.params['x2'] - (-0.3)) < 0.1
        
    def test_simple_fe_regression(self):
        """Test regression with single fixed effect."""
        # Generate test data with fixed effects
        np.random.seed(123)
        n = 200
        n_groups = 10
        
        data = pd.DataFrame({
            'group': np.random.randint(0, n_groups, n),
            'x': np.random.normal(0, 1, n),
        })
        
        # Add group fixed effects
        group_effects = np.random.normal(0, 1, n_groups)
        data['alpha'] = data['group'].map(dict(enumerate(group_effects)))
        data['y'] = data['alpha'] + 0.8 * data['x'] + np.random.normal(0, 0.2, n)
        
        # Estimate with fixed effects
        results = reghdfe(
            data=data,
            y='y',
            x='x',
            fe='group'
        )
        
        # Check results
        assert len(results.params) == 1
        assert 'x' in results.params.index
        assert results.fe_info['names'] == ['group']
        assert abs(results.params['x'] - 0.8) < 0.1
        
    def test_cluster_robust_se(self):
        """Test cluster-robust standard errors."""
        # Generate clustered data
        np.random.seed(456)
        n = 150
        n_clusters = 15
        
        data = pd.DataFrame({
            'cluster': np.random.randint(0, n_clusters, n),
            'x': np.random.normal(0, 1, n),
        })
        
        # Add cluster correlation
        cluster_effects = np.random.normal(0, 0.5, n_clusters)
        data['cluster_effect'] = data['cluster'].map(dict(enumerate(cluster_effects)))
        data['y'] = 0.6 * data['x'] + data['cluster_effect'] + np.random.normal(0, 0.3, n)
        
        # Estimate with cluster-robust SE
        results = reghdfe(
            data=data,
            y='y',
            x='x',
            cluster='cluster',
            cov_type='cluster'
        )
        
        # Check that clustering info is stored
        assert results.cluster_info is not None
        assert '1-way' in results.cluster_info['method']
        assert len(results.cluster_info['n_clusters']) == 1
        assert results.cluster_info['n_clusters'][0] == n_clusters
        
    def test_weights(self):
        """Test weighted regression."""
        np.random.seed(789)
        n = 100
        
        data = pd.DataFrame({
            'y': np.random.normal(1, 1, n),
            'x': np.random.normal(0, 1, n),
            'weight': np.random.uniform(0.5, 2.0, n)
        })
        
        # True relationship
        data['y'] = 0.7 * data['x'] + np.random.normal(0, 0.2, n)
        
        # Estimate with weights
        results = reghdfe(
            data=data,
            y='y',
            x='x',
            weights='weight'
        )
        
        assert results.weights_info is not None
        assert results.weights_info['variable'] == 'weight'
        assert abs(results.params['x'] - 0.7) < 0.2
        
    def test_summary_output(self):
        """Test that summary output is generated correctly."""
        np.random.seed(101)
        n = 50
        
        data = pd.DataFrame({
            'y': np.random.normal(0, 1, n),
            'x1': np.random.normal(0, 1, n),
            'x2': np.random.normal(0, 1, n),
            'fe': np.random.randint(0, 5, n)
        })
        
        data['y'] = 0.4 * data['x1'] - 0.2 * data['x2'] + np.random.normal(0, 0.1, n)
        
        results = reghdfe(
            data=data,
            y='y',
            x=['x1', 'x2'],
            fe='fe'
        )
        
        summary = results.summary()
        
        # Check that summary contains expected elements
        assert 'HDFE Linear regression' in summary
        assert 'Number of obs' in summary
        assert 'R-squared' in summary
        assert 'Within R-sq' in summary
        assert 'x1' in summary
        assert 'x2' in summary
        assert 'Fixed effects:' in summary
        
    def test_error_handling(self):
        """Test that appropriate errors are raised for invalid inputs."""
        data = pd.DataFrame({
            'y': [1, 2, 3],
            'x': [1, 2, 3]
        })
        
        # Test missing variable
        with pytest.raises(ValueError, match="Variables not found"):
            reghdfe(data=data, y='y', x='missing_var')
            
        # Test empty data
        with pytest.raises(ValueError):
            reghdfe(data=pd.DataFrame(), y='y', x='x')
            
        # Test invalid cov_type
        with pytest.raises(ValueError, match="cluster variables must be specified"):
            reghdfe(data=data, y='y', x='x', cov_type='cluster')


class TestDataGeneration:
    """Helper methods for generating test data."""
    
    @staticmethod
    def generate_panel_data(n_units=50, n_time=10, n_x_vars=3, seed=42):
        """Generate balanced panel data with unit and time fixed effects."""
        np.random.seed(seed)
        
        # Create panel structure
        units = np.repeat(range(n_units), n_time)
        time = np.tile(range(n_time), n_units)
        n_obs = n_units * n_time
        
        # Generate regressors
        X = np.random.normal(0, 1, (n_obs, n_x_vars))
        
        # True coefficients
        beta = np.array([0.5, -0.3, 0.8])
        
        # Unit and time fixed effects
        unit_fe = np.random.normal(0, 1, n_units)
        time_fe = np.random.normal(0, 0.5, n_time)
        
        # Generate dependent variable
        y = (X @ beta + 
             unit_fe[units] + 
             time_fe[time] + 
             np.random.normal(0, 0.2, n_obs))
        
        # Create DataFrame
        data = pd.DataFrame({
            'y': y,
            'unit': units,
            'time': time
        })
        
        for i in range(n_x_vars):
            data[f'x{i+1}'] = X[:, i]
            
        return data, beta
        
    def test_panel_data_estimation(self):
        """Test estimation on panel data with two-way fixed effects."""
        data, true_beta = self.generate_panel_data(n_units=30, n_time=8, n_x_vars=2)
        
        results = reghdfe(
            data=data,
            y='y',
            x=['x1', 'x2'],
            fe=['unit', 'time']
        )
        
        # Check that we recover true coefficients reasonably well
        assert abs(results.params['x1'] - true_beta[0]) < 0.1
        assert abs(results.params['x2'] - true_beta[1]) < 0.1
        
        # Check fixed effects info
        assert len(results.fe_info['names']) == 2
        assert 'unit' in results.fe_info['names']
        assert 'time' in results.fe_info['names']